package com.ibm.sip.test;

import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
import javax.servlet.sip.ServletParseException;
import javax.servlet.sip.SipFactory;
import javax.servlet.sip.SipServletRequest;
import javax.servlet.sip.SipServletResponse;
import javax.servlet.sip.SipApplicationSession;
import javax.servlet.sip.SipServlet;
import javax.servlet.sip.SipSession;
import javax.servlet.sip.SipURI;
import javax.servlet.sip.URI;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.ResourceBundle;

import com.sirf.studio.geoagent.*;
import com.sirf.studio.app.context.*;
import com.sirf.studio.geoagent.impl.*;
import com.sirf.studio.geoagent.factory.*;
import com.sirf.studio.app.*;
import com.sirf.studio.server.*;
import com.sirf.studio.geoservices.*;
import com.sirf.studio.ProviderInfo;
import com.sirf.studio.ProviderManager;
import com.sirf.studio.location.*;
import com.sirf.studio.infrastructure.logger.SSLogger;
import com.sirf.studio.infrastructure.logger.SSCategory;
import com.sirf.studio.context.*;
import com.sirf.studio.geoagent.exception.*;

/**
 * Servlet implementation class for Servlet: DemoAdapterSIP
 * 
 */
public class DemoAdapterSIP extends javax.servlet.sip.SipServlet implements
		javax.servlet.Servlet {

	private static final long serialVersionUID = 1L;
	public static final String EVENT_HEADER = "Event";
	public static final String PRESENCE_EVENT_PACKAGE = "presence";
	public static final String PIDF_FORMAT = "application/pidf+xml";
	private SipFactory _sipFactory;
	private DemoAdapterListener listener = null;
	private String publish_ETag = null;
	private boolean isUnpublish = false;
	private String tupleId = "1";
	private String PS_URI = null;

	/*
	 * Handles aquisition of the sip factory object
	 * @see javax.servlet.GenericServlet#init()
	 * 
	 */
	public void init() throws ServletException {
		System.out.println("DemoAdapterSIP init");
		ServletContext context = getServletContext();
		_sipFactory = (SipFactory) context.getAttribute(SIP_FACTORY);
		if (_sipFactory == null) {
			throw new ServletException("No SipFactory in context");
		}
		// On app startup, load zone definitions from properties file
		loadZoneDefinitions();
		// Register listener with AgentProvider 
		registerListenerForAlerts();

		
		ResourceBundle adapterBundle = ResourceBundle.getBundle("com.ibm.sip.test.DemoAdapter");
		//Properties file contains Presence Server URI -- to be updated for demo environment
		PS_URI = adapterBundle.getString("PS_URI");
	}

	/**
	 * Load country information from properties file
	 */
	private void loadZoneDefinitions() {
		GeoAgentFactory _geoagentfac = GeoAgentFactory.getInstance();
		ResourceBundle adapterBundle = ResourceBundle.getBundle("com.ibm.sip.test.DemoAdapter");
		String appID = "test_ad~test_ad";
		AgentProvisioningProvider provider = _geoagentfac.getAgentProvisioningProvider();
		GeocodingServiceProvider geocoder = (GeocodingServiceProvider) ProviderManager.connectToServiceProvider(null, ProviderManager.GEOCODING,appID);
		AddressInfo addr = new AddressInfo();
		// Address of St.Regis Hotel, location of Location 2.0 Summit
		addr.setField(AddressInfo.STREET, "125 3rd Street");
		addr.setField(AddressInfo.CITY, "San Francisco");
		addr.setField(AddressInfo.STATE, "CA");
		addr.setField(AddressInfo.COUNTRY_CODE, "US");
		Coordinates coord = null;
		try {
			Enumeration result = geocoder.geocode(addr, false, null);
			if (result.hasMoreElements()) {
				Landmark lm = (Landmark) result.nextElement();
				coord = lm.getQualifiedCoordinates();
			}
		} catch (Exception e1) {
			System.out.println("Failed to obtain coordinates for address "
					+ e1.toString());
		}

		if (coord != null) {
			String zoneKey = null;
			String[] zoneArgs = null;
			int radius;
			float direction;
			Enumeration enum_ = adapterBundle.getKeys();
			while (enum_.hasMoreElements()) {
				zoneKey = (String) (enum_.nextElement());
				if (zoneKey.contains("Zone")) {
					zoneArgs = adapterBundle.getString(zoneKey).split(",");
					radius = new Integer(zoneArgs[0]).intValue();
					direction = Float.parseFloat(zoneArgs[1]);
					try {
						CircleGeographicArea circleArea = new CircleGeographicArea(
								coord, radius, direction);
						GeoFence fence = _geoagentfac.createGeoFence("Fence#1");
						fence.setFenceName(zoneKey);
						fence.setFenceId(zoneKey);
						fence.setGeoBoundary(circleArea);
						boolean flag = provider.addFence(fence);
						if (!flag)
							System.out.println("Failed to create fence "
									+ zoneKey);
						else
							System.out.println("Fence " + zoneKey
									+ " created successfully");
					}

					catch (Exception e1) {
						System.out.println("Failed to create geographic area "
								+ e1.toString());
					}
				}
			}
		}
	}

	private void registerListenerForAlerts() {
		System.out.println("registerListenerForAlerts()");
		DemoAdapterListener listener = DemoAdapterListener.getInstance();
		listener.setAdapter(this);
		/* 	temp - uncomment for demo environment 
		GeoAgentFactory _fac = GeoAgentFactory.getInstance(); 
		AgentProvider provider = _fac.getAgentProvider(); 
		provider.registerAlertListener(listener);
		*/
	}

	public void publishAlert(String alertUser, String alertMessage) {
		String state = "open";
		createPublishRequest(state, alertUser, alertMessage);
	}

	public void createPublishRequest(String state, String user, String message) {
		System.out.println("DemoAdapterSIP: createPublishRequest called, user="
				+ user + " message=" + message);
		if (_sipFactory == null) {
			ServletContext sct = getServletContext();
			_sipFactory = (SipFactory) sct.getAttribute(SipServlet.SIP_FACTORY);
		}
		URI fromAddress = null;
		URI toAddress = null;
		try {
			fromAddress = _sipFactory.createURI("sip:" + user + "@demo.com");
			toAddress = _sipFactory.createURI("sip:" + user + "@demo.com");
		} catch (ServletParseException e1) {
			System.out.println("failed to create uri");
			return;
		}

		SipApplicationSession appSession = _sipFactory.createApplicationSession();
		SipServletRequest publishReq = _sipFactory.createRequest(appSession,"PUBLISH", fromAddress, toAddress);
		try {
			publishReq.setRequestURI(/* PS URI */_sipFactory.createURI(PS_URI));
		} catch (ServletParseException e1) {
			System.out.println("failed to create request uri");
			return;
		}

		// this is a presence event (header)
		publishReq.setHeader(EVENT_HEADER, PRESENCE_EVENT_PACKAGE);
		// set expires header
		if (!isUnpublish) {
			// expire in configured number fo seconds
			publishReq.setExpires(3600);
		} else {
			// expire now
			publishReq.setExpires(0);
		}
		// is this a refresh? then indicate so
		if (publish_ETag != null) {
			publishReq.addHeader("Sip-If-Match", publish_ETag);
		}

		String content = null;
		if (!isUnpublish) {
			content = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> <pidf:presence entity=\""
					+ uri2PresenceId(toAddress)
					+ "\" xmlns:pidf=\"urn:ietf:params:xml:ns:pidf\" xmlns:rpid=\"urn:ietf:params:xml:ns:pidf:rpid\" xmlns:dm=\"urn:ietf:params:xml:ns:pidf:data-model\"> <pidf:tuple id=\""
					+ tupleId
					+ "\"> <pidf:status> <pidf:basic>"
					+ state
					+ "</pidf:basic> </pidf:status> <note>"
					+ message
					+ " </note> </pidf:tuple> </pidf:presence>";
		}
		try {
			publishReq.setContent(content, PIDF_FORMAT);
		} catch (java.io.UnsupportedEncodingException e) {
			System.out.println("unable to publish due to exception!");
			return;
		}
		publishReq.setContentLength(isUnpublish ? 0 : content.length());
		System.out.println("sending PUBLISH (or unPUBLISH) request="
				+ publishReq);

		try {
			publishReq.send();
		} catch (IOException ex) {
			System.out.println("###DemoAdapterSIP: ERROR Unable to send PUBLISH message");
			System.out.println("###DemoAdapterSIP: ERROR PUBLISH message contents:\n"+ publishReq.toString());
			System.out.println("###DemoAdapterSIP: ERROR Exception:\n" + ex);
		}
	}

	/**
	 * Generate a String representation of the sip uri
	 * 
	 * @param uri
	 * @return the presence id in lowercase
	 */
	public static String uri2PresenceId(URI uri) {
		String res = null;

		if (uri.isSipURI()) {
			StringBuffer presenceId = new StringBuffer(((SipURI) uri).getUser()
					.length()
					+ ((SipURI) uri).getHost().length() + 7);
			presenceId.append("\"sip:");
			presenceId.append(((SipURI) uri).getUser());
			presenceId.append('@');
			presenceId.append(((SipURI) uri).getHost());
			res = presenceId.toString().toLowerCase();
		} else {
			res = uri.toString().toLowerCase();
		}

		return res;
	}

	/**
	 * In the response to the PUBLISH, look for the header "SIP-ETag" so we can
	 * update the PUBLISH state later.
	 */
	protected void doSuccessResponse(SipServletResponse response)
			throws ServletException, IOException {

		if ("PUBLISH".equals(response.getMethod())) {
			String tag = response.getHeader("SIP-ETag");

			if (tag != null) {
				log("found SIP-ETag: " + tag);
				// use the "to: field to figure out who was publishing
				String to = response.getHeader("To");
				log("To " + to);
				// clear stored info!
				if (to != null) {
					int startIndex = to.indexOf("<");
					int endIndex = to.indexOf(">", startIndex + 1);
					String address = to.substring(startIndex + 1, endIndex);
					log("Found " + address);
					publish_ETag = tag;
				}
			}
		}
		super.doSuccessResponse(response);
	}

	/**
	 * In response to an error
	 */
	protected void doErrorResponse(SipServletResponse response)
			throws ServletException, IOException {

		System.out.println("doErrorResponse message = " + response);
		super.doErrorResponse(response);
	}
}